use std::collections::HashMap;
use std::io;
use std::time::{Duration, Instant};

use crossterm::event;
use crossterm::event::{Event, KeyCode, MouseButton, MouseEventKind};
use tui::backend::Backend;
use tui::layout::Rect;
use tui::style::*;
use tui::widgets::*;
use tui::{Frame, Terminal};

use sea_battle_backend::data::*;

use crate::constants::*;
use crate::ui_screens::utils::{centered_rect_size, centered_text};
use crate::ui_screens::ScreenResult;
use crate::ui_widgets::game_map_widget::{ColoredCells, GameMapWidget};

struct SetBotsLayoutScreen {
    curr_boat: usize,
    layout: BoatsLayout,
}

type CoordinatesMapper = HashMap<Coordinates, Coordinates>;

pub fn set_boat_layout<B: Backend>(
    rules: &GameRules,
    terminal: &mut Terminal<B>,
) -> io::Result<ScreenResult<BoatsLayout>> {
    let mut model = SetBotsLayoutScreen {
        curr_boat: 0,
        layout: BoatsLayout::gen_random_for_rules(rules)
            .expect("Failed to generate initial boats layout"),
    };

    let mut coordinates_mapper = CoordinatesMapper::default();

    let mut last_tick = Instant::now();
    let mut is_moving_boat = false;
    loop {
        terminal.draw(|f| coordinates_mapper = ui(f, &mut model, rules))?;

        let timeout = TICK_RATE
            .checked_sub(last_tick.elapsed())
            .unwrap_or_else(|| Duration::from_secs(0));

        if event::poll(timeout)? {
            let mut move_boat = None;

            let event = event::read()?;
            if let Event::Key(key) = &event {
                match key.code {
                    KeyCode::Char('q') => return Ok(ScreenResult::Canceled),

                    // Select next boat
                    KeyCode::Char('n') => model.curr_boat += model.layout.number_of_boats() - 1,

                    // Rotate boat
                    KeyCode::Char('r') => {
                        model.layout.0[model.curr_boat].direction =
                            match model.layout.0[model.curr_boat].direction {
                                BoatDirection::Right => BoatDirection::Down,
                                _ => BoatDirection::Right,
                            }
                    }

                    // Move boat
                    KeyCode::Left => move_boat = Some((-1, 0)),
                    KeyCode::Right => move_boat = Some((1, 0)),
                    KeyCode::Up => move_boat = Some((0, -1)),
                    KeyCode::Down => move_boat = Some((0, 1)),

                    // Submit configuration
                    KeyCode::Enter => {
                        if model.layout.is_valid(rules) {
                            return Ok(ScreenResult::Ok(model.layout));
                        }
                    }

                    _ => {}
                }

                model.curr_boat %= model.layout.number_of_boats();

                // Apply boat move
                if let Some((x, y)) = move_boat {
                    let new_pos = model.layout.0[model.curr_boat].start.add_x(x).add_y(y);
                    if new_pos.is_valid(rules) {
                        model.layout.0[model.curr_boat].start = new_pos;
                    }
                }
            }
            // Mouse event
            else if let Event::Mouse(mouse) = event {
                let src_pos = Coordinates::new(mouse.column, mouse.row);

                // Start mouse action
                if MouseEventKind::Down(MouseButton::Left) == mouse.kind {
                    is_moving_boat = if let Some(pos) = coordinates_mapper.get(&src_pos) {
                        if let Some(b) = model.layout.find_boat_at_position(*pos) {
                            model.curr_boat = model.layout.0.iter().position(|s| s == b).unwrap();
                        }

                        true
                    } else {
                        false
                    }
                }
                // Handle continue mouse action
                else if is_moving_boat {
                    if let Some(pos) = coordinates_mapper.get(&src_pos) {
                        model.layout.0[model.curr_boat].start = *pos;
                    }

                    if let MouseEventKind::Up(_) = mouse.kind {
                        is_moving_boat = false;
                    }
                }
            }
        }

        if last_tick.elapsed() >= TICK_RATE {
            last_tick = Instant::now();
        }
    }
}

fn ui<B: Backend>(
    f: &mut Frame<B>,
    model: &mut SetBotsLayoutScreen,
    rules: &GameRules,
) -> CoordinatesMapper {
    let errors = model.layout.errors(rules);

    // Color of current boat
    let current_boat = ColoredCells {
        color: Color::Green,
        cells: model.layout.0[model.curr_boat].all_coordinates(),
    };

    // Color of invalid boats
    let mut invalid_coordinates = vec![];
    for (idx, pos) in model.layout.boats().iter().enumerate() {
        if idx == model.curr_boat {
            continue;
        }

        if !model
            .layout
            .check_present_boat_position(idx, rules)
            .is_empty()
        {
            invalid_coordinates.append(&mut pos.all_coordinates());
        }
    }
    let invalid_boats = ColoredCells {
        color: Color::Red,
        cells: invalid_coordinates,
    };

    // Color of other boats
    let mut other_boats_cells = vec![];
    for boat in &model.layout.0 {
        other_boats_cells.append(&mut boat.all_coordinates());
    }

    let other_boats = ColoredCells {
        color: Color::Gray,
        cells: other_boats_cells,
    };

    let mut coordinates_mapper = HashMap::new();

    let mut legend = "n        next boat     \n\
                     r        rotate boat    \n\n\
                     ← ↓↑ →   move boat      \n\n"
        .to_string();
    if errors.is_empty() {
        legend.push_str("Enter    confirm layout");
    }

    let mut game_map_widget = GameMapWidget::new(rules)
        .set_default_empty_char(' ')
        .add_colored_cells(current_boat)
        .add_colored_cells(invalid_boats)
        .add_colored_cells(other_boats)
        .set_title("Choose your boat layout")
        .set_yield_func(|c, r| {
            for i in 0..r.width {
                for j in 0..r.height {
                    coordinates_mapper.insert(Coordinates::new(r.x + i, r.y + j), c);
                }
            }
        })
        .set_legend(legend);

    // Color of neighbors if boats can not touch
    if !rules.boats_can_touch {
        let mut boats_neighbors_cells = vec![];
        for boat in &model.layout.0 {
            for pos in boat.neighbor_coordinates(rules) {
                boats_neighbors_cells.push(pos);
            }
        }

        game_map_widget = game_map_widget.add_colored_cells(ColoredCells {
            color: Color::Rgb(30, 30, 30),
            cells: boats_neighbors_cells,
        });
    }

    let (w, h) = game_map_widget.estimated_size();
    let area = centered_rect_size(w, h, &f.size());
    f.render_widget(game_map_widget, area);

    if !errors.is_empty() {
        let messages = ["INVALID_LAYOUT", errors[0]];
        for (i, msg) in messages.iter().enumerate() {
            let paragraph = Paragraph::new(*msg).style(Style::default().fg(Color::Red));
            f.render_widget(
                paragraph,
                centered_text(
                    msg,
                    &Rect::new(f.size().x, area.bottom() + i as u16, f.size().width, 1),
                ),
            );
        }
    }

    coordinates_mapper
}
